from typing import Any

class BindLocation: ...

class TCPSocket(BindLocation):
    bind_addr: Any
    def __init__(self, address, port) -> None: ...

class UnixSocket(BindLocation):
    bind_addr: Any
    def __init__(self, path) -> None: ...

class AbstractSocket(BindLocation):
    bind_addr: Any
    def __init__(self, abstract_socket) -> None: ...

class Application:
    @classmethod
    def resolve(cls, full_path): ...
    wsgi_app: Any
    def __init__(self, wsgi_app) -> None: ...
    def server_args(self, parsed_args): ...
    def server(self, parsed_args): ...

class GatewayYo:
    gateway: Any
    def __init__(self, gateway) -> None: ...
    def server(self, parsed_args): ...

def parse_wsgi_bind_location(bind_addr_string: str): ...
def parse_wsgi_bind_addr(bind_addr_string: str): ...
def main() -> None: ...
